/*******************************************************************************
*                                                                              *
*    Leibniz Institute on Aging - Fritz Lipmann Institute                      *
*    Genomic Ananlysis                                                         *
*                                                                              *
*    niels jahn	                                                               *
*    2014,2015                                                                 *
*                                                                              *
*******************************************************************************/

#include <fstream>
#include <stdexcept>
#include <sstream>

#include "search_around_class.h"

using namespace std;

search_around_class::search_around_class(size_t b,size_t a)
:before(b),after(a),data(),own_data(false),req_len(0),hits()
{
}

void search_around_class::load_data_from_file(const string &fn)
{	if (own_data)
	{	delete data;
	}
	own_data=true;
	data=new list<string>();

	ifstream	in(fn.c_str());
	if (!in || !in.is_open())
	{	throw runtime_error("search_around_class::search_around_class : can't open '"+fn+"' for reading");
	}
	string	line;
	while (true)
	{	getline(in,line);
		if (in.eof())
		{	break;
		}
		data->push_back(line);
	}	
}

void search_around_class::use_data(list<string> &data_foreign)
{	if (own_data)
	{	delete data;
	}
	own_data=false;
	data=&data_foreign;
}	

void search_around_class::validate_pattern(const string &p)
{	if (string::npos!=p.find_first_not_of("ACGT:"))
	{	throw runtime_error("search_around_class::validate_pattern : unhandled char in pattern '"+p+"'");
	}
	size_t pos=p.find(':');
	if (pos!=string::npos)
	{	if (pos==p.size()-1)
		{	return;
		}
		string	pattern=p.substr(0,pos);
		string	replace=p.substr(pos+1);
		if (pattern.size()!=replace.size())
		{	throw runtime_error("search_around_class::validate_pattern : pattern size and replace size differs '"+p+"'");
		}
	}
}

void search_around_class::search(const string &combined_pattern)
{	hits.clear();
	all_count_hits=0;
	hits_per_read_histogram.clear();

	validate_pattern(combined_pattern);
	string	pattern;
	string	replace;
	size_t pos=combined_pattern.find(':');
	if (pos==string::npos)
	{	pattern=combined_pattern;
	}
	else
	{	pattern=combined_pattern.substr(0,pos);
		if (pos<combined_pattern.size()-1)
		{	replace=combined_pattern.substr(pos+1);
		}
	}	
	
	for (list<string>::iterator i=data->begin(); i!=data->end(); ++i)
	{	pos=0;
		int count_hits=0;

		while (string::npos!=(pos=i->find(pattern,pos)))
		{	string	out;
			if (pos<before)
			{	out=i->substr(0,pos);
			}
			else
			{	out=i->substr(pos-before,before);
			}
			out+=i->substr(pos,pattern.size()+after);
			hits[out].count++;

			if (replace.size() && hits[out].replace.size()==0)
			{	string	rep;
				if (pos<before)
				{	rep=i->substr(0,pos);
				}
				else
				{	rep=i->substr(pos-before,before);
				}
				rep+=replace;
				rep+=i->substr(pos+pattern.size(),after);
				hits[out].replace=rep;
			}
			pos++;
			count_hits++;
		}
		all_count_hits+=count_hits;
		hits_per_read_histogram[count_hits]++;
	}

	// search replacements
	if (replace.size())
	{	for (list<string>::iterator i=data->begin(); i!=data->end(); ++i)
		for (map<string,hit>::iterator j=hits.begin(); j!=hits.end(); ++j)
		{	pos=0;
			while (string::npos!=(pos=i->find(j->second.replace,pos)))
			{	j->second.replace_count++;
				pos++;
			}
		}
	}

	req_len=before+after+pattern.size();
}

void search_around_class::output(ostream &o,bool pattern_hit_info,int max_hits,bool do_sort) const
{	// put results into list for sorting
	list<pair<string,hit> >	sdata;
	for (map<string,hit>::const_iterator i=hits.begin(); i!=hits.end(); ++i)
	{	sdata.push_back(pair<string,hit>(i->first,i->second));
	}

	if (do_sort)
	{	sdata.sort(hit_sorter());
	}

	int sum=0;
	list<string>	ret;

	for (list<pair<string,hit> >::const_iterator i=sdata.begin(); i!=sdata.end(); ++i)
	{	ostringstream out;
		out << i->first << '\t';
		out << i->second.count;
		if (i->second.replace.size())
		{	out << '\t' << i->second.replace << '\t' << i->second.replace_count;
		}
		ret.push_back(out.str());
		if (i->first.size()==req_len)
		{       sum+=i->second.count;
		}
	}
	if (pattern_hit_info)
	{	o << endl;
		o << all_count_hits << " pattern hit(s) in " << data->size() <<  " reads; " << double(all_count_hits)/double(data->size()) << " hits/per read" << endl;
		o << "hits per read\tcount" << endl;
		for (map<int,int>::const_iterator  i=hits_per_read_histogram.begin(); i!=hits_per_read_histogram.end(); ++i)
		{	o << i->first << "\t\t" << i->second << endl;
		}
		o << endl;
	}
	int shown=0;
	for (list<string>::iterator i=ret.begin(); i!=ret.end(); ++i)
	{       o << *i << endl;
		shown++;
		if (max_hits && shown>=max_hits && do_sort)
		{	break;
		}
	}
}

