/*******************************************************************************
*                                                                              *
*    Leibniz Institute on Aging - Fritz Lipmann Institute                      *
*    Genomic Ananlysis                                                         *
*                                                                              *
*    niels jahn	                                                               *
*    2014,2015                                                                 *
*                                                                              *
*******************************************************************************/

#include "tag_primer_amplicon.h"
#include <ctype.h>
#include "expand_iupac.h"
#include <sstream>
#include <iostream>
#include <iomanip>
#include <cmath>
#include <set>

#include <boost/lexical_cast.hpp>

using namespace std;
using namespace boost;

ostream& operator << (ostream &o,const tag_primer_amplicon &t)
{	o << t.primer_f;
	o << t.primer_r;
	o << t.amplicon;
	o << t.tag;
	return o;
}	

void tag_primer_amplicon::toupper()
{	primer_f.toupper();
	primer_r.toupper();
	amplicon.toupper();
	tag.toupper();
}

void tag_primer_amplicon::create_pattern()
{	pattern_f=expand_iupac(primer_f.get_data());
	pattern_r=expand_iupac(primer_r.get_data());

}	

void tag_primer_amplicon::create_iupac()
{	iupac_f=iupac(primer_f.get_data());
	iupac_r=iupac(primer_r.get_data());
}	

bool tag_primer_amplicon::match_primers(const string &p1,const string &p2)
{	if (pattern_f.size()==0 || pattern_r.size()==0)
	{	create_pattern();
	}
	int tag_size=tag.get_data().size();
	string t(tag.get_data());
	if (tag_size && ( p1.substr(0,tag_size)!=t || p2.substr(0,tag_size)!=t))
	{
		return false;
	}
//cout << "match_primers:" << endl << p1 << endl << p2 << endl;	
	for (list<string>::iterator i=pattern_f.begin(); i!=pattern_f.end(); ++i)
	{	if (p1.substr(tag_size,i->size())==*i)
		{	for (list<string>::iterator j=pattern_r.begin(); j!=pattern_r.end(); ++j)
			{	if (p2.substr(tag_size,j->size())==*j)
				{	
					return true;
				}
			}
		}	
		if (p2.substr(tag_size,i->size())==*i)
		{	for (list<string>::iterator j=pattern_r.begin(); j!=pattern_r.end(); ++j)
			{       if (p1.substr(tag_size,j->size())==*j)
				{	
					return true;
				}
			}
		}
	}	
	return false;
}

bool tag_primer_amplicon::match_primers_iupac(const string &p1,const string &p2)
{	if (iupac_f.size()==0 || iupac_r.size()==0)
	{	create_iupac();
	}
	int tag_size=tag.get_data().size();
	string t(tag.get_data());
	if (tag_size && ( p1.substr(0,tag_size)!=t || p2.substr(0,tag_size)!=t))
	{	return false;
	}
	if (match_iupac(p1.substr(tag_size,primer_f.size()),iupac_f) 
		&& match_iupac(p2.substr(tag_size,primer_r.size()),iupac_r))
	{	return true;
	}
	if (match_iupac(p2.substr(tag_size,primer_f.size()),iupac_f)
		&& match_iupac(p1.substr(tag_size,primer_r.size()),iupac_r))
	{	return true;
	}
	return false;
}	

bool tag_primer_amplicon::match_primers_iupac_single(const std::string &p1,const std::string &p2)
{	if (iupac_f.size()==0 || iupac_r.size()==0)
	{	create_iupac();
	}
	int tag_size=tag.get_data().size();
	string t(tag.get_data());
	if (tag_size && ( p1.substr(0,tag_size)!=t || p2.substr(0,tag_size)!=t))
	{	return false;
	}
	if (match_iupac(p1.substr(tag_size,primer_f.size()),iupac_f) 
		|| match_iupac(p2.substr(tag_size,primer_r.size()),iupac_r))
	{	return true;
	}
	if (match_iupac(p2.substr(tag_size,primer_f.size()),iupac_f)
		|| match_iupac(p1.substr(tag_size,primer_r.size()),iupac_r))
	{	return true;
	}
	return false;
}

string tag_primer_amplicon::info() const
{	ostringstream ret;
	ret << bucket << '\t' << hits << '\t' << tag.get_data() << '\t' << primer_f.get_name() << '\t' << primer_r.get_name();
	return ret.str();
}	

bool tag_primer_amplicon::operator < (const tag_primer_amplicon &t)
{	if (primer_f.get_name()==t.primer_f.get_name())
	{	if (primer_r.get_name()==t.primer_r.get_name())
		{	return tag.get_name()<t.tag.get_name();
		}
		else
		{	return primer_r.get_name()<t.primer_r.get_name();
		}
	}
	else
	{	return primer_f.get_name()<t.primer_f.get_name();
	}
}	

void all_tag_primer_amplicon::add(const string &s)
{	fasta	f;
	f.load(s);
	tag_primer_amplicon	tpa;
	list<fasta_piece>	tags;
	for (unsigned int i=0; i<f.size(); ++i)
	{	switch (i)
		{	case 0 :
				tpa.primer_f=f[0];
				break;
			case 1 :
				tpa.primer_r=f[1];
				break;
			case 2 :
				tpa.amplicon=f[2];
				break;
			default :
				if (f[i].get_name()=="INTERNAL_MOTIF")
				{	tpa.internal_motif=f[i].get_data();
				}
				else
				{	tags.push_back(f[i]);
				}
				break;
		}
	}	
	if (tags.size()==0)
	{	data.push_back(tpa);
	}
	else
	{	for (list<fasta_piece>::iterator i=tags.begin(); i!=tags.end(); ++i)
		{	tag_primer_amplicon	n_tpa(tpa);
			n_tpa.tag=*i;
			string	tag_string(n_tpa.tag.get_name());
			size_t pos=tag_string.find("bucket_id=");
			if (pos!=string::npos)
			{	n_tpa.bucket=lexical_cast<int>(tag_string.substr(pos+10));
				tag_string.erase(pos);
				n_tpa.tag.set_name(tag_string);
			}	
			data.push_back(n_tpa);
		}
	}	
//	data.sort();
	int index=1;
	for (list<tag_primer_amplicon>::iterator i=data.begin(); i!=data.end(); ++i)
	{	if (i->bucket==0)
		{	i->bucket=index++;
		}
		else
		{	index=i->bucket+1;
		}
	}
	// verify bucket numbers only used once
	set<int>	used;
	for (list<tag_primer_amplicon>::iterator i=data.begin(); i!=data.end(); ++i)
	{	if (used.find(i->bucket)!=used.end())
		{	ostringstream n_str;
			n_str << i->bucket;
			throw runtime_error("all_tag_primer_amplicon::add() : bucket number used twice"+n_str.str());
		}
		used.insert(i->bucket);
	}	
}

void all_tag_primer_amplicon::toupper()
{	for (list<tag_primer_amplicon>::iterator i=data.begin(); i!=data.end(); ++i)
	{	i->toupper();
	}
}
	
bool all_tag_primer_amplicon::check_tag_size(int x)	
// tag size of 0 is accepted
{	for (list<tag_primer_amplicon>::iterator i=data.begin(); i!=data.end(); ++i)
	{	if (i->tag.size()!=0 && i->tag.size()!=x)
		{	return false;
		}
	}
	return true;
}	

int all_tag_primer_amplicon::get_max_bucket_id() const
{	int ret=0;
	for (list<tag_primer_amplicon>::const_iterator i=data.begin(); i!=data.end(); ++i)
	{	if (i->bucket>ret)
		{	ret=i->bucket;
		}
	}
	return ret;
}	

void all_tag_primer_amplicon::write_amplicons(const string &dir) const
{	for (list<tag_primer_amplicon>::const_iterator i=data.begin(); i!=data.end(); ++i)
	{	ostringstream o;
		o << dir << "/ampl_bucket_" << setfill('0') << setw(3) << i->get_bucket() << ".fa";
		ofstream out(o.str().c_str());
		if (!out || !out.is_open())
		{	throw runtime_error("all_tag_primer_amplicon::write_amplicons : can't open '"+o.str()+"'");
		}
		out << ">" << i->tag.get_data();
		out << "_" << i->amplicon.get_name();
		out << endl;
		out << i->tag.get_data() << i->amplicon.get_data();
		fasta f(i->tag);
		f.revcomp();
		out << f.get_data();
		out << endl;
	}
}	
		 

