/*******************************************************************************
*                                                                              *
*    Leibniz Institute on Aging - Fritz Lipmann Institute                      *
*    Genomic Ananlysis                                                         *
*                                                                              *
*    niels jahn	                                                               *
*    2014,2015                                                                 *
*                                                                              *
*******************************************************************************/

#include "fasta.h"
#include "match_iupac.h"
#include <list>

class tag_primer_amplicon
{	
	void create_pattern();
	void create_iupac();

	int bucket;
	public:

	tag_primer_amplicon()
	:bucket(0),tag(),primer_f(),primer_r(),amplicon(),iupac_f(),iupac_r(),hits(0),internal_motif()
	{	
	}

	fasta_piece	tag;
	fasta_piece	primer_f;
	fasta_piece	primer_r;
	fasta_piece	amplicon;
	iupac		iupac_f;
	iupac		iupac_r;
	int hits;
	std::string	internal_motif;

	std::list<std::string>	pattern_f;
	std::list<std::string>	pattern_r;

	void toupper();
	bool match_primers(const std::string&,const std::string&);
	bool match_primers_iupac(const std::string&,const std::string&);
	bool match_primers_iupac_single(const std::string&,const std::string&);

	int get_bucket() const
	{	return bucket;
	}

	std::string info() const;

	bool operator < (const tag_primer_amplicon&);

	friend class all_tag_primer_amplicon;
};

std::ostream& operator << (std::ostream&,const tag_primer_amplicon&);

class all_tag_primer_amplicon
{	std::list<tag_primer_amplicon>	data;

	public:
	all_tag_primer_amplicon()
	:data()
	{
	}

	void add(const std::string&);

	void toupper();

	bool check_tag_size(int);

	int get_max_bucket_id() const;

	size_t size() const
	{	return data.size();
	}

	void write_amplicons(const std::string&) const;

	class iterator
	{	std::list<tag_primer_amplicon>::iterator i;
		
		public:

		iterator(std::list<tag_primer_amplicon>::iterator j)
		{	i=j;
		}

		tag_primer_amplicon& operator * ()
		{	return *i;
		}
		tag_primer_amplicon* operator -> ()
		{	return &(*i);
		}
		bool operator == (const iterator &j)
		{	return i==j.i;
		}
		bool operator != (const iterator &j)
		{	return i!=j.i;
		}
		const iterator& operator ++ ()
		{	++i;
			return *this;
		}
	};	

	iterator begin()
	{	return iterator(data.begin());
	}
	iterator end()
	{	return iterator(data.end());
	}

};


